/*
 * Copyright (c) 1999 4 Crazy Men All Rights Reserved
 *
 * Modifications :
 *       21/05/1999 : - debut du programme
 *
 *       09/06/1999 : - Modification d'un petit bug d'affichage :
 *                      Certain carres sortaient du cadre en haut et 
 *                      a droite de l'image
 *       13/06/1999 : - les carres representant l'image recue et l'image affichee
 *                      sont de forme differente (la couleur differente n'etait
 *                      pas suffisante : impression noir et blanc).
 *
 */
package video.affichage;


import java.util.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.color.*;
import java.awt.geom.*;
import java.awt.event.*;

/**
 * Cette fenetre est chargee de representer graphiquement
 * les statistiques creees par le buffer de playout
 * 
 * @version 0.1 21 Mai 1999
 * @author BONNIN Francis
 */
public class FenetreStats extends Frame implements ActionListener{

  /** hauteur de la fenetre  */
  private int haut = 500;

  /** largeur de la fenetre */
  private int larg = 700;

  /** l'ancienne hauteur de la fenetre  */
  int oldWidth = haut;

  /** l'ancienne largeur de la fenetre */
  int oldHeight = larg;

  /** l'image affichee */
  BufferedImage image;

  /** les statistiques a afficher */
  StatistiquesPlayout stats;

  /** constructeur
   * @param _can le canvas (avec buffeur de playout)
   */
  public FenetreStats(StatistiquesPlayout _stats) {
    stats = _stats;
    setSize(larg, haut);
    calculeImage(larg, haut);

    Button bouttonQuit = new Button("Fermer");
    Panel paneButton = new Panel();
 
    bouttonQuit.addActionListener(this);

    setLayout(new BorderLayout());
    paneButton.add(bouttonQuit);

    add("Center", new graph());
    add("South", paneButton);
    
    setTitle("Statistiques");
    show();
  }



  /** effectue les actions correspondantes aux boutons de l'interface
   */
  public void actionPerformed(ActionEvent e) {
    String bouton = e.getActionCommand();
    if (bouton == "Fermer"){
      dispose();
    }
  }


  /** 
   * cree l'image representative avec les dimensions donnees
   * @param larg La largeur de l'image
   * @param haut L'hauteur de l'image
   */
  public void calculeImage(int larg, int haut) {
    int hauteurTicks = 5;
    int distanceBordure = 20;
    String ch;
    FontMetrics fontMetrics;
		Rectangle2D rect;

    image = new BufferedImage(larg, haut, ColorSpace.TYPE_RGB);
    Graphics2D gc2D = image.createGraphics();


		//met le couleur de fond
    gc2D.setColor(Color.white);
    gc2D.fillRect(0, 0, larg, haut);


		//affiche les axes
    gc2D.setColor(Color.black);
    gc2D.drawLine(distanceBordure ,haut  - distanceBordure,distanceBordure ,distanceBordure);
    gc2D.drawLine(distanceBordure ,haut - distanceBordure ,larg - distanceBordure ,haut - distanceBordure);

		//affiche la fleche en y
    gc2D.drawLine(distanceBordure ,distanceBordure - hauteurTicks, distanceBordure - hauteurTicks,distanceBordure );
    gc2D.drawLine(distanceBordure ,distanceBordure - hauteurTicks, distanceBordure + hauteurTicks,distanceBordure);
    gc2D.drawLine(distanceBordure - hauteurTicks,distanceBordure , distanceBordure + hauteurTicks,distanceBordure );

		//affiche la fleche en x
    gc2D.drawLine(larg - distanceBordure + hauteurTicks,haut - distanceBordure, larg - distanceBordure ,haut - distanceBordure - hauteurTicks);
    gc2D.drawLine(larg - distanceBordure + hauteurTicks,haut - distanceBordure, larg - distanceBordure ,haut - distanceBordure + hauteurTicks);
    gc2D.drawLine(larg - distanceBordure  ,haut - distanceBordure - hauteurTicks, larg - distanceBordure  ,haut - distanceBordure + hauteurTicks);


		//affichage de la legende
		ch = "Numero de paquet";
		fontMetrics = gc2D.getFontMetrics();
		rect = fontMetrics.getStringBounds(ch, gc2D);
		gc2D.drawString(ch, distanceBordure + (int)fontMetrics.getStringBounds("10000", gc2D).getWidth()  + hauteurTicks , (int)rect.getHeight()); 
		
		ch = "Temps en millisecondes";
		fontMetrics = gc2D.getFontMetrics();
		rect = fontMetrics.getStringBounds(ch, gc2D);
		gc2D.drawString(ch, larg - (int)rect.getWidth() - distanceBordure, haut - distanceBordure - hauteurTicks - (int)(hauteurTicks/2)); 
		



		//taille des max
    long mx = -1;
    long my = -1;
		CoordonneesPlayout elem;
		Vector statReception = stats.getStatReception();
		Vector statAffichage = stats.getStatAffichage();

		for (int i = 0; i < statReception.size() ; i++){
      elem = (CoordonneesPlayout)statReception.elementAt(i);
      if (elem.gety() > my)
        my = elem.gety();
    }
    
		for (int i = 0; i < statReception.size() ; i++){
      elem = (CoordonneesPlayout)statReception.elementAt(i);
      if (elem.getx() > mx)
        mx = elem.getx();
    }
    
		for (int i = 0; i < statAffichage.size() ; i++){
      elem = (CoordonneesPlayout)statAffichage.elementAt(i);
      if (elem.getx() > mx)
        mx = elem.getx();
    }
    
    
		if (mx < 1)
				mx = -1;
		if (my < 1)
				my = -1;

		//taille des carres
	  int tailleCarre;
    if (my > 0)
      tailleCarre = (int)((haut - 2*distanceBordure) / (my + 1));
    else
      tailleCarre = 6;

		if (mx != -1)
				mx = mx + ((mx * tailleCarre) / (larg - 2*distanceBordure));  
		if (my != -1)
				my = my + 1;  

		if (tailleCarre < 6)
			 tailleCarre = 6;




		//affichage des unites de l'axe x
    long nbPixelsTiks = 10;
    long nbTiks = larg / nbPixelsTiks;
    long intervalleTicksX = mx / nbTiks;
    long intervalleCoordX = intervalleTicksX * 10;
    int positX;
    for(long x = 0; x <= mx; x = x + intervalleTicksX){
      positX = distanceBordure + (int)((x * (larg - 2*distanceBordure)) / mx);
      if ((x % intervalleCoordX) == 0){
        ch = (new Long(x)).toString();
        fontMetrics = gc2D.getFontMetrics();
				rect = fontMetrics.getStringBounds(ch, gc2D);

        gc2D.drawString(ch, positX - (int)(rect.getWidth() / 2), haut - distanceBordure  + 2 + (int)(rect.getHeight())); 
        gc2D.drawLine(positX ,haut - distanceBordure ,positX , haut  - distanceBordure - hauteurTicks - hauteurTicks/2);
      }
      else if ((x % intervalleTicksX) == 0){
        gc2D.drawLine(positX ,haut - distanceBordure ,positX , haut  - distanceBordure - hauteurTicks );
      }
    }


		//affichage des unites de l'axe y
    long intervalleTicksY = 10;
    long intervalleCoordY = 50;
		int positY;
		
    for(long y = 1; y <= my; y++){
				positY =  haut - distanceBordure - (int)((y * (haut - 2*distanceBordure)) / my);
				if ((y % intervalleCoordY) == 0){
						ch = (new Long(y)).toString();
				 fontMetrics = gc2D.getFontMetrics();
						rect = fontMetrics.getStringBounds(ch, gc2D);

        gc2D.drawString(ch, distanceBordure + hauteurTicks, positY + (int)(rect.getHeight() / 2)); 
      }
      if ((y % intervalleTicksY) == 0){
        gc2D.drawLine(distanceBordure  ,positY,distanceBordure + hauteurTicks ,positY);
      }
    }

		//affiche le max en x
		gc2D.setPaintMode();
		if (mx != -1){
        ch = (new Long(mx)).toString();
        fontMetrics = gc2D.getFontMetrics();
				rect = fontMetrics.getStringBounds(ch, gc2D);
				gc2D.setColor(Color.white);
				gc2D.fillRect(larg - (int)(rect.getWidth()), haut - distanceBordure  + hauteurTicks +2 + (int)(rect.getHeight()) - (int)rect.getHeight(), (int)rect.getWidth(), (int)rect.getHeight() );
				gc2D.setColor(Color.green);
				gc2D.drawRect(larg - (int)(rect.getWidth()) - 1, haut - distanceBordure  + hauteurTicks +2 + (int)(rect.getHeight()) - (int)rect.getHeight() - 1, (int)rect.getWidth() + 1,  (int)rect.getHeight() + 1 );
				gc2D.setColor(Color.black);
        gc2D.drawString(ch, larg - (int)(rect.getWidth()), haut - distanceBordure  + hauteurTicks +2 + (int)(rect.getHeight())); 
		}
		//affiche le max en y
		if (my != -1){
        ch = (new Long(my)).toString();
        fontMetrics = gc2D.getFontMetrics();
				rect = fontMetrics.getStringBounds(ch, gc2D);
				gc2D.setColor(Color.white);
				gc2D.fillRect( distanceBordure + hauteurTicks , distanceBordure - (int)rect.getHeight(), (int)rect.getWidth(), (int)rect.getHeight() );
				gc2D.setColor(Color.green);
				gc2D.drawRect( distanceBordure + hauteurTicks - 1, distanceBordure - (int)rect.getHeight() - 1, (int)rect.getWidth() + 1, (int)rect.getHeight() + 1 );
				gc2D.setColor(Color.black);
        gc2D.drawString(ch, distanceBordure + hauteurTicks , distanceBordure); 
		}

		int tailleCarre2;

		//affiche les statistiques de reception
    gc2D.setColor(Color.blue);
		for (int i = 0; i < statReception.size() ; i++){
				elem = (CoordonneesPlayout)statReception.elementAt(i);
	      positY =  haut - distanceBordure - (int)((elem.gety() * (haut - 2*distanceBordure)) / my);
				positX = distanceBordure + (int)((elem.getx() * (larg - 2*distanceBordure)) / mx);
				gc2D.drawRoundRect(positX, positY-tailleCarre,tailleCarre , tailleCarre, tailleCarre/3, tailleCarre/3);
				tailleCarre2 = tailleCarre/2;
				gc2D.fillRect(positX + (tailleCarre/2) - (tailleCarre2/2), positY - (tailleCarre/2) - (tailleCarre2/2),tailleCarre2 , tailleCarre2);

		}

		//affiche les statistiques d'affichage
    gc2D.setColor(Color.red);
		for (int i = 0; i < statAffichage.size() ; i++){
				elem = (CoordonneesPlayout)statAffichage.elementAt(i);
	      positY =  haut - distanceBordure - (int)((elem.gety() * (haut - 2*distanceBordure)) / my);
				positX = distanceBordure + (int)((elem.getx() * (larg - 2*distanceBordure)) / mx);
				gc2D.drawRect(positX, positY-tailleCarre,tailleCarre , tailleCarre);
				tailleCarre2 = tailleCarre/3;
				gc2D.fillRect(positX + (tailleCarre/2) - (tailleCarre2/2), positY - (tailleCarre/2) - (tailleCarre2/2),tailleCarre2 , tailleCarre2);
        gc2D.drawLine(positX ,positY-tailleCarre, positX + tailleCarre, positY);
        gc2D.drawLine(positX ,positY, positX + tailleCarre, positY-tailleCarre);
		}
  }


  private class graph extends Canvas {
    
    /** constructeur
     */
    public graph() {
    }
    
    /** 
     * redefinition de update pour eviter le scintillement 
     */
    public void update(Graphics gc) {
      paint(gc);
    }
    
    
    /** 
     * affichage du canvas
     */
    public void paint(Graphics gc) {
      Dimension d = getSize();
      int maxWidth = d.width;
      int maxHeight = d.height;
      
      if ((oldWidth != maxWidth) || (oldHeight != maxHeight)){
        // on recalcule l'image
        calculeImage(maxWidth, maxHeight);
        oldWidth = maxWidth;
        oldHeight = maxHeight;
      }
      
      gc.drawImage(image, 0, 0, null);
    }
  }

}
