/*
 * Copyright (c) 1999 4 Crazy Men All Rights Reserved
 * 
 * Modifications :
 *      19/05/1999 : - ajout de la possibilite de redimentionner ou non l'image
 *                   - modification de l'update : on ne recoit plus que des
 *                     NotifyObject
 *
 *      20/05/1999 : - generation de statistiques
 *                     (vecteurs de CoordonneerPlayout)
 *
 *      08/06/1999 : - plus d'attente active dans le run.
 *                     run n'est lance que lorsque on connait 
 *                     variationDelai et delaiEstime (donnees utiles pour le
 *                     calcul de la duree du playout)
 *      09/06/1999 : - on ne passe plus un objet charge d'afficher les messages de
 *                     fonctionnement (ou autres) mais on notifie ces messages
 *                   - on peut maintenant faire un reset au milieu d'une reception
 *                     ce qui permet de recalculer le delai du buffer de playout au 
 *                     milieu d'une session video (apres un tres long retard par 
 *                     exemple
 *                   - bug repare, il causait par exemple des duplications d'images 
 *                     dans le buffer de playout meme. Il fallait mettre update en 
 *                     synchronized. 
 *      10/06/1999 : - avant on generait tout le temps des statistiques, ce qui 
 *                     pouvait utiliser de la memoire pour rien. Maintenant, par
 *                     defaut on ne genere aucune statistiques
 *                     on notifie un booleen (GENERE_STATISTIQUES) pour indiquer 
 *                     qu'on veut les generer
 *                   - augmentation de la priorite du thread ('ajoutPriorite') charge
 *                     de l'affichage pour de la videoConference c'est important non ? 
 *                   - On peut maintenant commencer a partir de n'importe quel numero
 *                     d'image (source d'un bug quand on fesait le reset
 *       16/06/1999 : - On ferme maintenant explicitement une session video : 
 *                      La derniere image ne contient plus un booleen a true 
 *                      indiquant la fin.
 *       17/06/1999 : - les statistiques n'etaient pas bonnes dans le cas
 *                      ou le nombre maximal de d'images affichees et recues 
 *                      n'etaient pas egales.
 *       18/06/1999 : - bug corrige : pb quand on envoyait tout le temps les infos 
 *                      delai estime et variation.
 *
 *
 */
package video.affichage;

import video.tools.*;

import java.util.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.color.*;


/**
 * Ce canvas s'occupe d'afficher les images qu'on lui envoie. Il 
 * implemente un buffer de playout, ce qui lui permet de pallier
 * aux problemes de delai, redondance, entre les differentes images
 * de la video
 * emmet des informations (messages, debug)
 * produit des statistiques sur demande.
 *
 * @version 0.1 19 Mai 1999
 * @author BONNIN Francis
 * @see Afficheur
 */
public class FrameDisplay extends Canvas
  implements Observer, Runnable {

  /* -------------------------------CONSTANTES-----------------------------  */

  /** constante representant l'envoi d'un Objet PlayoutMessage. */
  public static final int PLAYOUT_MESSAGE = 10;
  
  /** constante representant l'envoi d'un Objet videoImage. */
  public static final int VIDEO_IMAGE = 0;
  
  /** constante representant l'information DELAI_ESTIME (l'info est un Long). */
  public static final int DELAI_ESTIME = 1;

  /**constante representant l'information VARIATION_DELAI (l'info est un Long). */
  public static final int VARIATION_DELAI = 2;
  
  /** constante representant l'information INITIALISE_SESSION_VIDEO
   * (l'info est inutilisee). */
  public static final int INITIALISE_SESSION_VIDEO = 3;
  
  /** constante representant l'information INITIALISE_SESSION_VIDEO
   * (l'info est inutilisee). */
  public static final int RESET_SESSION_VIDEO = 4;
  
  /**constante representant l'information INITIALISE_SESSION_VIDEO
   * (l'info est inutilisee).*/
  public static final int CLOSE_SESSION_VIDEO = 6;
  
  /** constante representant l'information COEFFICIENT_VARIATION_DELAI
   * (l'info est un Double).*/
  public static final int COEFFICIENT_VARIATION_DELAI = 7;
  
  /** constante representant l'information COEFFICIENT_VARIATION_DELAI
   * (l'info est un Boolean). */
  public static final int ZOOM_IMAGE  = 8;

  /** constante representant l'information GENERE_STATISTIQUES
   * (l'info est un Boolean). */
  public static final int GENERE_STATISTIQUES  = 9;

  /* ------------------------------------------------------------  */

  /** Objet permettant au FrameDisplay d'etre observable */
  public SimpleObservable application;


  /** Le coefficient multiplicateur pour la variation de delai (entre 0 et 4). */
  private double coefficientVariationDelai = 1; 

  /** le delai estime (temps ecoule entre l'emission et la reception)
   *  doit etre pracise par l'appel a setDelaiEstime. */
  private long delaiEstime = -1;
  
  /** La variation de delai estimee. */
  private long variationDelai = -1; 
  
  /** thread charge d'actualiser l'affichage des images. */
  private Thread actualiseurImage;

  /** Le buffer de playout. */
  private Vector playout = new Vector();

  /** La derniere image recue. */
  private VideoImage derniereImage = new VideoImage(null, -777, 100, 0);
  
  /**  Le numero de l'image qui devrait etre affichee. */
  private int numeroImageNormalementAffichee = -777;
  
  /** L'image qui est actuellement affichee */
  private BufferedImage imageCourante;
  
  /**la duree du playout en millisecondes */
  private long dureePlayout; 
  
  /** Derniere frequence d'affichage de la video en cours*/
  private long dernierDelai;
  
  /** moment ou commence la session video */
  private long tempsDebutSession = 0;
  
  /** moment ou commence la reception d'images */
  private long tempsDebutReception = 0;
  
  /** moment ou finit l'affichage d'images */
  private long tempsFinAffichage = 0;
  
  /** definit l'activite du thread d'affichage (inverse par rapport a la logique */
  private boolean threadActif = true;
  
  /** definit si le taille de l'image video doit etre
   * ajustee a l'espace disponible. */
  private boolean zoomImage = false;
  
  /** indique si on doit effacer l'imqge */
  private boolean effaceImage = false;

  /** utilite : booleen permettant de savoir si on recoit la premiere image s'une session */
  private boolean etudieDebutSession;
  
  /** utilite : booleen permettant de savoir si on recoit la premiere image d'un groupe*/
  private boolean etudieDebutReception;
  
  /** Les statistiques a la reception des images */
  private Vector statReception = new Vector();
  
  /** Booleen indiquant si on doit generer des statistiques */
  private boolean genereStatistiques = false;
  
  /** Les statistiques a l'affichage des images */
  private Vector statAffichage = new Vector();
  
  /** l'increment de priorite du thread d'affichage */
  private int ajoutPriorite = 2;

  /** Booleen indiquant si on a deja commence a afficher les images */
  private boolean affichageEnCours = false;

  /** l'ancienne hauteur de la fenetre  */
  int oldWidth = 0;

  /** l'ancienne largeur de la fenetre */
  int oldHeight = 0;

  /* ------------------------------------------------------------  */
  
    
  /** 
   * Constructeurs de l'objet
   */	
  public FrameDisplay() {
    application = new SimpleObservable();
    initConnection();
  }


  private void notifye(int num, String str){
    application.notify(new NotifyObject(PLAYOUT_MESSAGE,
                                        new PlayoutMessage(num,str),
                                        this));
  }
  

  /** 
   * Si il est possible de commencer l'affichage de la video
   * (c'est a dire que toutes les informations pour l'affichage 
   * sont disponible)
   * alors on commence l'affichage
   */	
  private void tesTNSetDebutAffichage(){
    if ((variationDelai != -1) && (delaiEstime != -1)) {
      //lancement du thread si necessaire

      if (!actualiseurImage.isAlive()){ 
        actualiseurImage.start();
      }
    }
  }

  /** 
   * donne au buffer de playout la valeur du delai estime
   *
   * @param _delaiEstime le delai estime
   */	
  public void setDelaiEstime(long _delaiEstime){
    notifye(3, "Donnee 'delaiEstime' recue");
    delaiEstime = _delaiEstime;
    tesTNSetDebutAffichage();
  }
  
  /** 
   * donne au buffer de playout la valeur de la variation de delai
   * @param _variationDelai la variation de delai
   */	
  public void setVariationDelai (long _variationDelai){
    notifye(3, "Donnee 'variationDelai' recue");
    variationDelai = _variationDelai;
    tesTNSetDebutAffichage();
  }
  
  /** 
   * definit si le taille de l'image video doit
   * etre ajustee a l'espace disponible
   *
   * @param _zoomImage le booleen le definissant
   */	
  public void setZoomImage(boolean _zoomImage){
    zoomImage = _zoomImage;
    effaceImage = true;
    repaint(0);
  }
  
  /** 
   * definit si on doit ou non generer des statistiques
   *
   * @param _genereStatistiques le booleen le definissant
   */	
  public void setGenereStatistiques(boolean _genereStatistiques){
    genereStatistiques = _genereStatistiques;
  }
  /** 
   * donne au buffer de playout la valeur du delai estime
   * @param _delaiEstime le delai estime
   */	
  public void setCoefficientVariationDelai(double _coefficientVariationDelai){
    coefficientVariationDelai = _coefficientVariationDelai;
  }
  
  /** 
   *  Permet d'initialiser ou reinitialiser une session video
   *  appelee au tout debut
   */	
  public void initConnection(){
    statReception = new Vector();
    statAffichage = new Vector();
    tempsDebutSession = 0;
    etudieDebutSession = true;
    tempsFinAffichage = 0;
    imageCourante = new BufferedImage(getSize().width + 1 ,
                                      getSize().height + 1,
                                      ColorSpace.TYPE_CMYK);
    repaint(10);
    dernierDelai = 100;
    resetConnection();
  }


  /** 
   *  Permet de continuer une session video
   *  Ne remet pas les statistiques a zero
   *  Appellee pour recalculer le delai
   */	
  public void resetConnection(){
    delaiEstime = -1;
    variationDelai = -1;
    killThread();
    actualiseurImage = new Thread(this);
		actualiseurImage.setPriority(Math.min(actualiseurImage.getPriority() + ajoutPriorite, Thread.MAX_PRIORITY));
    playout = new Vector();
    derniereImage = new VideoImage(null, -777, 100, 0);
    numeroImageNormalementAffichee = -777;
    tempsDebutReception = 0;
    etudieDebutReception = true;
  }

  /** 
   *  Permet de terminer une session video
   */	
  public void closeConnection(){
    killThread();
    tempsFinAffichage = System.currentTimeMillis();
  } 
  
  public void killThread(){
    if ((actualiseurImage != null) && (actualiseurImage.isAlive())){
      threadActif = false;   
      while (threadActif = false){
        try {
          actualiseurImage.sleep(10);
        } 
        catch(InterruptedException e) {
          System.err.println("Erreur dans le sleep();");
          e.printStackTrace();
        }
      }
    }
  }
  
  /** 
   *  Permet de recuperer les statistiques du buffer de playout
   *  Cette methode devrait etre appellee apres une session video
   *  (apres un appel a closeConnection())
   */	
  public StatistiquesPlayout getStats(){
    return new StatistiquesPlayout(((tempsFinAffichage == 0)?System.currentTimeMillis():tempsFinAffichage) - 
                                   ((tempsDebutSession == 0)?System.currentTimeMillis():tempsDebutSession),
                                   Math.max(statAffichage.size(),statReception.size()) + 
                                   (derniereImage.getNum() - statAffichage.size()),
                                   statReception,
                                   statAffichage);



  } 

  /** 
   * Methode run du thread charge d'actualiser l'affichage des images
   */
  public void run() {
    long tempsTmp;
    long frequenceEmission;
    long tempsPreced;
    long dureeAttendue;

    dureePlayout = (delaiEstime +
                    (new Double(variationDelai * coefficientVariationDelai)).longValue());
    notifye(3, "DUREE DU PLAYOUT : " + transformeTemps(dureePlayout));

    if (tempsDebutReception == 0)
      dureeAttendue = 0;
    else 
      dureeAttendue = (System.currentTimeMillis() - tempsDebutReception);

    notifye(3, "DUREE ATTENDUE : " + transformeTemps(dureeAttendue));

    dureePlayout = dureePlayout - dureeAttendue;
    if (dureePlayout < 0)
      dureePlayout = 0;

    notifye(3, "DUREE RESTANTE : " + transformeTemps(dureePlayout));
    
    try {
      actualiseurImage.sleep(dureePlayout);
    } 
    catch(InterruptedException e) {
      System.err.println("Erreur dans le sleep();");
      e.printStackTrace();
    }
    affichageEnCours = true;   
    while(threadActif) {
      tempsPreced = System.currentTimeMillis();
      frequenceEmission = afficheNouvelleImage();
      tempsTmp = frequenceEmission - 
        (System.currentTimeMillis() - tempsPreced);

      if (tempsTmp < 0){
        notifye(1, "'");
      }
      else {
        try {
          actualiseurImage.sleep(tempsTmp);
        } 
        catch(InterruptedException e) {
          System.err.println("Erreur dans le sleep();");
          e.printStackTrace();
        }
      }			
    }
    affichageEnCours = false;
    threadActif = true;
  }
  
  /** 
   * Methode qui observe un evenement de type nouvelle VideoImage ou
   * information
   *
   * @see VideoImage
   */
  public synchronized void update(Observable observable, Object o) {
    NotifyObject no = (NotifyObject)o;
    switch (no.type) {
    case DELAI_ESTIME :
      setDelaiEstime(((Long)(no.obj)).longValue());
      break;
    case VARIATION_DELAI :
      setVariationDelai(((Long)(no.obj)).longValue());
      break;
    case INITIALISE_SESSION_VIDEO :
      initConnection();
      break;
    case RESET_SESSION_VIDEO :
      resetConnection();
      break;
    case CLOSE_SESSION_VIDEO :
      closeConnection();
      break;
    case COEFFICIENT_VARIATION_DELAI :
      setCoefficientVariationDelai(((Double)(no.obj)).doubleValue());
      break;
    case ZOOM_IMAGE :
      setZoomImage(((Boolean)(no.obj)).booleanValue());
      break;
    case GENERE_STATISTIQUES :
      setGenereStatistiques(((Boolean)(no.obj)).booleanValue());
      break;
    case VIDEO_IMAGE:
      VideoImage vim = (VideoImage)(no.obj);

      /* Si l'image est de qualite nulle  : on l'ignore
       * Elle est probableme
       */
      if (vim.getQualite() <= 0){
        notifye(3, "Image de qualite nulle recue : ignoree ! ");
        break;
      }

      /* Initialisation des temps de depart
       */
      if (etudieDebutReception) {
        tempsDebutReception = System.currentTimeMillis();
        etudieDebutReception = false;
        notifye(3, "Debut Reception Images : " + vim.getNum() + "( " + transformeTemps(tempsDebutReception) + " )");
        
      }
      if (etudieDebutSession) {
        tempsDebutSession = System.currentTimeMillis();
        etudieDebutSession = false;
      }
			if (genereStatistiques)
				statReception.add(new CoordonneesPlayout(System.currentTimeMillis() -
                                                 tempsDebutSession,
                                                 vim.getNum()));
      
      VideoImage premiereImage;
      synchronized (playout) {
        if (playout.size() >= 1)
          premiereImage = (VideoImage)playout.firstElement();
        else
          premiereImage = null;
      }
      
      /* on n'a pas encore commence a afficher les images
       * il se peut que l'image recue soit la une des premieres images arrivees en
       * retard
       */
      if (((derniereImage.getNum()) != -777) && 
          (affichageEnCours == false) && 
          (premiereImage != null) &&
          (vim.getNum() < premiereImage.getNum())){
        
        /* on ajoute la premiere image et on comble
         */
        synchronized (playout) {
          if (((VideoImage)playout.firstElement()).getNum() == premiereImage.getNum()){
            for (int i = premiereImage.getNum() - 1; i > vim.getNum(); i--) {
              playout.insertElementAt(new VideoImage(null, i, vim.getDelay(), 0), 0);
            }	  
            notifye(3, "L' image n " + vim.getNum() + " est peut etre la premiere de la serie");
            playout.insertElementAt(vim, 0);
            numeroImageNormalementAffichee = vim.getNum() -1;
          }
        }
      }
      else {

        if (numeroImageNormalementAffichee >= vim.getNum()){
          /* l'image arrive trop tard : elle aurait deja du etre affichee
           */
						//if  (vim.getNum() > (derniereImage.getNum()))
						//		derniereImage = vim;
          notifye(3, "Image n " + vim.getNum() +
                  " (de qualite " + vim.getQualite() +
                  ") arrivee trop tard (Image deja Affichee)");
         }
        else {
        
        if  (((derniereImage.getNum()) == -777) || (vim.getNum() == ((derniereImage.getNum()) + 1))){
          if ((derniereImage.getNum()) == -777)
            numeroImageNormalementAffichee = vim.getNum() -1;
          /* l'image recue est bien la suivante : il n'y a pas eu de perte
           */
          playout.add(vim);
          derniereImage = vim;
        }
        else {
          if  (vim.getNum() <= (derniereImage.getNum())){
            /* l'image recue avais pris du retard : il y a eu inversion
             *  il se peut aussi qu'on soit en presence d'une duplication
             *  de paquet
             *  il faut retrouver l'indice de l'image dans le vecteur et 
             *  remplacer l'image (de moins bonne qualite) par celle-ci.
             *  cette partie doit etre en section critique
             */
            synchronized (playout) {
              int indice_image = playout.size() - 1 -
                (derniereImage.getNum() - vim.getNum());
              if (indice_image >= 0) {
                /* l'image n'avait pas encore ete affichee
                 *  on peut donc la remplacer
                 *  sinon, c'est trop tard, l'image de substitution
                 *  a deja ete affichee 
                 */						
                if (vim.getQualite() > ((VideoImage)playout.elementAt(indice_image)). getQualite()){
                  playout.setElementAt(vim, indice_image);
                  notifye(3, "Image n " + vim.getNum() +
                          " (de qualite "+ vim.getQualite() + 
                          ") arrivee a temps ");
                }
                else {
                  /* une image de qualite superieure est deja en place
                   */
                  notifye(3, "Image n " + vim.getNum() +
                          " (de qualite "+ vim.getQualite() + 
                          ") inutile ");
                }
              }
              else {
                /*l'image est arrivee trop tard
                 */                
                notifye(3, "Image n " + vim.getNum() +
                        " (de qualite " + vim.getQualite() +
                        ") arrivee trop tard ");
              }
            }
          }
          else {
            /* il y a eu une ou plusieurs pertes			
             *  on doit reconstituer les images perdues
             *  on met a la place des images perdues des copies
             *  de la derniere image recue
             */
            for (int i = Math.max(numeroImageNormalementAffichee, derniereImage.getNum()) + 1; i < vim.getNum(); i++) {
              notifye(3, "Image n " + i + " en retard (Perte ??) ");
              playout.add(new VideoImage(null, i, vim.getDelay(), 0));
            }
	  
            playout.add(vim);
            derniereImage = vim;
          }
        }
        }
        break;
      }
    }
  }
  /** 
   *  est appelee pour afficher l'image en tete du buffer de playout
   *  retourne le delai d'attente avant d'afficher l'image suivante
   */
  public synchronized long afficheNouvelleImage(){
    numeroImageNormalementAffichee = numeroImageNormalementAffichee + 1;
		// notifye(3, "Affichage de l'images  --> " + numeroImageNormalementAffichee);
 
    if (playout.size() >= 1){
      /* il y a plusieurs images dans le buffer de playout, 
       *  on peut donc supprimer la premiere et ainsi l'afficher
       *  partie en section critique
       */
      try{
        VideoImage vim;
        synchronized (playout) {
          vim = (VideoImage)playout.remove(0);
        }
        notifye(1, "#");
        if (vim.getImage() == null){
          notifye(2, new Integer(vim.getNum()).toString() + "\t vide");
        }
        else {
          if (genereStatistiques)
            statAffichage.add(new CoordonneesPlayout(System.currentTimeMillis()
                                                     - tempsDebutSession,
                                                     vim.getNum()));
          notifye(2, new Integer(vim.getNum()).toString());
        }
        imageCourante = vim.getImage();
        //notifye(3, "Affichage reel         --> " + vim.getNum() + "( " + numeroImageNormalementAffichee + " )");
        repaint(10);
        dernierDelai = vim.getDelay();
        return dernierDelai;
      }
      catch (ArrayIndexOutOfBoundsException e){
        /* ne devrait normalement pas arriver car 
         *  on a verifie la taille du buffer avant
         */
        e.printStackTrace();
        return dernierDelai;
      }
    }
    else {
      /* il n'y a plus d'image dans le buffer de playout,
       *  on laisse l'ancienne image affichee
       *  le temps d'attente est probablement trop court.
       *  il y a peut etre aussi eu un pb reseau
       */
      notifye(1, "~");
      return dernierDelai;
    }
  }
	
  /** 
   * redefinition de update pour eviter le scintillement 
   */
  public void update(Graphics gc) {
    paint(gc);
  }

  /** 
   * affichage du canvas
   */
  public void paint(Graphics gc) {
    Dimension d = getSize();
    int maxWidth = d.width;
    int maxHeight = d.height;
    
    if (effaceImage || (oldWidth != maxWidth) || (oldHeight != maxHeight)){
      oldWidth = maxWidth;
      oldHeight = maxHeight;
      gc.drawImage(new BufferedImage(maxWidth + 1 ,
                                     maxHeight + 1,
                                    ColorSpace.TYPE_CMYK),
                   0, 0, null);
      effaceImage = false;
    }
    
    if (imageCourante != null) {
      if (zoomImage){
        gc.drawImage(imageCourante.getScaledInstance(d.width, d.height,
                                                     Image.SCALE_DEFAULT),
                     0, 0, null);
      }
      else{
        gc.drawImage(imageCourante, (maxWidth - imageCourante.getWidth()) / 2,
                     (maxHeight - imageCourante.getHeight()) / 2, null);
      }
    }
  }
  
  /** 
   * rend un temps affichable de maniere lisible
   * @param temps le temps (en millisecondes) a transformer
   * @return la chaine de caractere representant le nombre
   */
  public String transformeTemps(long temps) {
    String chaine = "";
    int seco = (int)(temps / 1000);
    int minu = (int)(seco / 60);
    seco = (int)(seco % 60);
    int heur = (int)(minu / 60);
    minu = (int)(minu % 60);    
    int mill = (int)(temps % 1000);
    return(heur + "h " + minu + "mn " + seco + "," + mill + "s");
  }
}
