/*
 * Copyright (c) 1999 4 Crazy Men All Rights Reserved
 * 
 * Modifications :
 *       21/05/1999 : - ajout de l'affichage graphique des statistiques generees
 *                      par le buffer de playout (@see StatistiquesPlayout)
 *
 *       08/06/1999 : - l'arret d'une session video en plein milieu n'oblige plus
 *                      le FrameDisplay a attendes toutes les images
 *                    - ajout d'une seconde textbox de messages
 *
 *       09/06/1999 : - Le playoutTester est maintenant un observeur de
 *                      MessagePlayout notifies par le FrameDisplay.
 *                      Il n'implemente plus la classe Affichage.
 *                    - ajout de la possiblite de faire une pause pendant 
 *                      l'envoi des images.(permet de tester le reset
 *                      du Buffer de Playout
 *                    - on peut maintenant preciser le nombre maximum de redondance
 *                      d'une image
 *                    - on envoyait une image en trop, c'est bete, non ?
 *       10/06/1999 : - Pendant la pause, on peut maintenant sauter des images 
 *                      (max et min)
 *       16/06/1999 : - On ferme maintenant explicitement une session video : 
 *                      La derniere image ne contient plus un booleen a true 
 *                      indiquant la fin.
 *                    - Ajout d'informations a modifier dans l'interface
 *
 */
package video.affichage;

import video.tools.*;

import java.util.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.color.*;
import java.awt.event.*;
import java.awt.geom.*;
/**
 * Cette classe permet de visualiser comment reagit le buffer de playout
 * en fonction de certains parametres
 * 
 * @version 0.1 19 Mai 1999
 * @author BONNIN Francis
 * @see Afficheur
 */

public class PlayoutTester extends Observable implements Observer, Runnable, ActionListener{
		
  /** La frequence d'emission des images en millisecondes */
		private long frequenceEmission = 100;

  /** La variation de delai possible entre la reception de chaque image en millisecondes */
  private long variationDelai = 600;

  /** Le pourcentage de pertes des paquets d'image */
  private long pourcentagePertes = 5;

  /** Le nombre d'images envoyees vers le reseau simule*/
  private int nbImagesEnvoyees = 50;

  /** Le delai  que met un paquet pour aller de la source vers la destination en millisecondes */
  private long delaiTransmission = 10;

  /** La taille en pixels du cote de l'image envoyee */
  private int tailleCoteImage = 50;

  /** Le nombre d'images apres lesquelle se fera l'envoi au buffer de playout du delai de transmission
   *  et de la variation de delai */
  private int nbImageAvtEmissionInfos = 1;

  /** Le coefficient multiplicateur de la variation du delai (calcul du delai de playout) */
  private double CoefficientVariationDelai = 1; 

  /** Le pourcentage de duplication des paquets d'image */
  private long pourcentageDuplication = 6;

  /** Le maximum de duplications d'une image */
  private int nbMaxDuplication = 3;

  /** Le nombre minimal d'images perdues apres une pause */
  private int NbMinImagesPerduesApresPause = 10;

  /** Le nombre maximal d'images perdues apres une pause */
  private int NbMaxImagesPerduesApresPause = 20;
  /* --------------------------------------------------------------------------------------------- */

  /** Le numero de la premiere  */
  private int numPremiereImage = 0;
  
  /** booleen indiquant si l'image doit etre zoomee ou non au depart*/
  private boolean zoomImage = false; 


  /* --------------------------------------------------------------------------------------------- */

  /** Le buffer de Playout (qui est un canvas) */
  private FrameDisplay displayImage;

  /** La VideoImage envoyee au buffer de playout*/
  private VideoImage vi;

  /** La fenetre graphique du 'Playout Tester' */
  private maFenetre fenetreTest;
		
  /** thread charge d'envoyer les images */
  private Thread emetteur;
		
  /** permet de savoir quand on doit envoyer les informations au buffer de playout */
  private int EmissionANum = -1;

  /** Synchronise les modifications de EmissionANum */
  private Integer EmissionANumSync = new Integer(0);

  /** booleen indiquant si on doit arreter l'envoyeur d'images */
  private boolean arretDemande = false;
		
  /** booleen indiquant si on met l'emission en pause */
  private boolean pauseEmission = false; 

  /* --------------------------------------------------------------------------------------------- */


  /** Constructeur de l'objet PlayoutTester
   */
  public PlayoutTester() {
    displayImage = new FrameDisplay();
    this.addObserver(displayImage);
    displayImage.application.addObserver(this);
    fenetreTest = new maFenetre (this);
    notifye(FrameDisplay.ZOOM_IMAGE, new Boolean(zoomImage));
    new FenetreVideoConf(displayImage,tailleCoteImage + 100,tailleCoteImage + 100);
    fenetreTest.setAff();
  }

  /** Affiche une Legende sur la sortie standard
   */
  private void afficheAide(){
    fenetreTest.afficheMessages2("\n");
    fenetreTest.afficheMessages2("|------------------------------------------------------------------------|");
    fenetreTest.afficheMessages2("| . : envoi d'une image                                                  |");
    fenetreTest.afficheMessages2("| # : image affichee                                                     |");
    fenetreTest.afficheMessages2("| ~ : plus d'images dans le buffer                                       |");
    fenetreTest.afficheMessages2("| ' : Warning: temps d'affichage plus long que le taux raffraichissement |");
    fenetreTest.afficheMessages2("|------------------------------------------------------------------------|");
    fenetreTest.afficheMessages2("\n");
  }

  /** Pour demarrer une session de test
   */
  private void demarrer(){
    arretDemande = false;
    emetteur = new Thread(this);
    emetteur.start();
  }

  /** Pour arreter une session de test
   */
  private void arreter(){
    arretDemande = true;
    notifye(FrameDisplay.CLOSE_SESSION_VIDEO,null);
    fenetreTest.setlabelStart("Demarrer");
  }


  /** effectue les actions correspondantes aux boutons de l'interface
   */
  public void actionPerformed(ActionEvent e) {
    String bouton = e.getActionCommand();
    if (bouton == "Fermer"){
      fenetreTest.dispose();
      System.exit(0);
    }
    if (bouton == "Demarrer"){
      demarrer();
    }
    if (bouton == "Arreter"){
      arreter();
    }
    if (bouton == "Afficher"){
      new FenetreStats(displayImage.getStats());
    } 
    if (bouton == " Pause Emission "){
      fenetreTest.setlabelPause("UnPause Emission");
      pauseEmission = true;
    } 
    if (bouton == "UnPause Emission"){
      fenetreTest.setlabelPause(" Pause Emission ");
      pauseEmission = false;
    } 
  }
  
 /** La fenetre graphique du 'Playout Tester' 
  *  Permet le test du buffer de Playout de maniere interactive
  */ 
  private class maFenetre extends Frame {
    private int larg = 650;
    private int haut = 580;
    private TextField tf1;
    private TextField tf2;
    private TextField tf3;
    private TextField tf4;
    private TextField tf5;
    private TextField tf6;
    private TextField tf7;
    private TextField tf8;
    private TextField tf9;
    private TextField tf10;
    private TextField tf11;
    private TextField tf12;
    private TextArea texteProducteur;
    private TextArea texteConsommateur;
    private TextArea texteMessages1;
    private TextArea texteMessages2;
    private Button bouttonQuit;
    private Button bouttonStart;
    private Button bouttonAff;
    private Button bouttonPause;

    private int tailleTexte = 0;


    /** Le constructeur
     * @param evt l'objet s'occupant de traiter les actions correspondantes aux boutons
     */ 
    public maFenetre(ActionListener evt) {
      bouttonQuit = new Button("Fermer");
      bouttonStart = new Button("Demarrer");
      bouttonAff = new Button("Afficher");
      bouttonPause = new Button(" Pause Emission ");
      texteProducteur = new TextArea(15,70);
      texteConsommateur = new TextArea(15,70);
      texteMessages1 = new TextArea(7,70);
      texteMessages2 = new TextArea(3,70);						
      tf1 = new TextField(6);
      tf2 = new TextField(6);
      tf3 = new TextField(6);
      tf4 = new TextField(6);
      tf5 = new TextField(6);
      tf6 = new TextField(6);
      tf7 = new TextField(6);
      tf8 = new TextField(6);
      tf9 = new TextField(6);
      tf10 = new TextField(6);
      tf11 = new TextField(6);
      tf12 = new TextField(6);
					
      Label lb1 = new Label("Frequence d'Emission (ms) :");
      Label lb2 = new Label("Max Variation Delai (ms) :");
      Label lb3 = new Label("Pourcentage de Pertes (%) :");
      Label lb4 = new Label("Nb Images Envoyees :");
      Label lb5 = new Label("Delai de Transmission (ms) :");
      Label lb6 = new Label("Taille Cote Image (pixels) :");
      Label lb7 = new Label("Nb images avant emission infos (>0):");
      Label lb8 = new Label("Coefficient Multiplicateur (0-4):");
      Label lb9 = new Label("Pourcentage Duplication (%) :");
      Label lb10 = new Label("Nb Max Duplication d'une Image :");
      Label lb11 = new Label("Nb Min Images perdues apres la pause :");
      Label lb12 = new Label("Nb Max Images perdues apres la pause :");

      bouttonQuit.addActionListener(evt);
      bouttonStart.addActionListener(evt);
      bouttonAff.addActionListener(evt);
      bouttonPause.addActionListener(evt);
	
      texteConsommateur.setEditable(false);
      texteProducteur.setEditable(false);
      texteMessages1.setEditable(false);
      texteMessages2.setEditable(false);
						
      Panel paneText = new Panel(new BorderLayout());
      Panel paneText2 = new Panel(new GridLayout(1, 2, 10, 3));
      Panel paneMessages = new Panel(new BorderLayout());
      Panel paneButton = new Panel();
      Panel paneSaisie = new Panel(new BorderLayout());
      Panel paneSaisiea = new Panel(new GridLayout(6, 1, 0, 0));
      Panel paneSaisieb = new Panel(new GridLayout(6, 1, 0, 0));
      Panel paneSaisie1a = new Panel(new FlowLayout(FlowLayout.RIGHT));
      Panel paneSaisie2a = new Panel(new FlowLayout(FlowLayout.RIGHT));
      Panel paneSaisie3a = new Panel(new FlowLayout(FlowLayout.RIGHT));
      Panel paneSaisie4a = new Panel(new FlowLayout(FlowLayout.RIGHT));
      Panel paneSaisie5a = new Panel(new FlowLayout(FlowLayout.RIGHT));
      Panel paneSaisie6a = new Panel(new FlowLayout(FlowLayout.RIGHT));
      Panel paneSaisie1b = new Panel(new FlowLayout(FlowLayout.RIGHT));
      Panel paneSaisie2b = new Panel(new FlowLayout(FlowLayout.RIGHT));
      Panel paneSaisie3b = new Panel(new FlowLayout(FlowLayout.RIGHT));
      Panel paneSaisie4b = new Panel(new FlowLayout(FlowLayout.RIGHT));
      Panel paneSaisie5b = new Panel(new FlowLayout(FlowLayout.RIGHT));
      Panel paneSaisie6b = new Panel(new FlowLayout(FlowLayout.RIGHT));

      setLayout(new BorderLayout());
      setTitle("Playout Tester");

			paneSaisie.add("West",paneSaisiea);						
			paneSaisie.add("East",paneSaisieb);						
      paneButton.add(bouttonStart);
      paneButton.add(bouttonPause);
      paneButton.add(bouttonAff);
      paneButton.add(bouttonQuit);

      paneSaisie1a.add(lb1);
      paneSaisie1a.add(tf1);
      paneSaisie1b.add(lb2);
      paneSaisie1b.add(tf2);

      paneSaisie2a.add(lb3);
      paneSaisie2a.add(tf3);
      paneSaisie2b.add(lb8);
      paneSaisie2b.add(tf8);

      paneSaisie3a.add(lb5);
      paneSaisie3a.add(tf5);
      paneSaisie3b.add(lb4);
      paneSaisie3b.add(tf4);

      paneSaisie4a.add(lb7);
      paneSaisie4a.add(tf7);
      paneSaisie4b.add(lb6);
      paneSaisie4b.add(tf6);

      paneSaisie5a.add(lb9);
      paneSaisie5a.add(tf9);
      paneSaisie5b.add(lb10);
      paneSaisie5b.add(tf10);

      paneSaisie6a.add(lb11);
      paneSaisie6a.add(tf11);
      paneSaisie6b.add(lb12);
      paneSaisie6b.add(tf12);


      paneSaisiea.add(paneSaisie4a);
      paneSaisiea.add(paneSaisie3a);
      paneSaisiea.add(paneSaisie2a);
      paneSaisiea.add(paneSaisie5a);
      paneSaisiea.add(paneSaisie6a);
      paneSaisiea.add(paneSaisie1a);
					
      paneSaisieb.add(paneSaisie4b);
      paneSaisieb.add(paneSaisie3b);
      paneSaisieb.add(paneSaisie2b);
      paneSaisieb.add(paneSaisie5b);
      paneSaisieb.add(paneSaisie6b);
      paneSaisieb.add(paneSaisie1b);
					
      paneMessages.add("North",texteMessages1);
      paneMessages.add("South",texteMessages2);

      paneText2.add(texteProducteur);
      paneText2.add(texteConsommateur);
      paneText.add("Center", paneText2);
      paneText.add("South", paneMessages);
						



      add("North",paneSaisie);
      add("Center", paneText);
      add("South", paneButton);
      this.pack();
      this.setVisible(true);
      setSize(larg, haut);
      show();
    }

    /** actualise les variables du Testeur de playout avec les valeurs donnees 
     *  par l'utilisateur (par l'intermediaire de l'interface)
     */ 
    public void getAff(){
      frequenceEmission = new Long(tf1.getText()).longValue();
      variationDelai = new Long(tf2.getText()).longValue();
      pourcentagePertes = new Long(tf3.getText()).longValue();
      nbImagesEnvoyees = new Integer(tf4.getText()).intValue();
      delaiTransmission = new Long(tf5.getText()).longValue();
      tailleCoteImage = new Integer(tf6.getText()).intValue();
      nbImageAvtEmissionInfos = new Integer(tf7.getText()).intValue();
      CoefficientVariationDelai = new Double(tf8.getText()).intValue();
      pourcentageDuplication = new Long(tf9.getText()).longValue();
      nbMaxDuplication = new Integer(tf10.getText()).intValue();
      NbMinImagesPerduesApresPause = new Integer(tf11.getText()).intValue();
      NbMaxImagesPerduesApresPause = new Integer(tf12.getText()).intValue();
    }
				
    /** actualise l'interface avec les variables du Testeur de playout
     */ 
    public void setAff(){
      tf1.setText((new Long(frequenceEmission)).toString());
      tf2.setText((new Long(variationDelai)).toString());
      tf3.setText((new Long(pourcentagePertes)).toString());
      tf4.setText((new Integer(nbImagesEnvoyees)).toString());
      tf5.setText((new Long(delaiTransmission)).toString());
      tf6.setText((new Integer(tailleCoteImage)).toString());
      tf7.setText((new Integer(nbImageAvtEmissionInfos)).toString());
      tf8.setText((new Double(CoefficientVariationDelai)).toString());
      tf9.setText((new Long(pourcentageDuplication)).toString());
      tf10.setText((new Integer(nbMaxDuplication)).toString());
      tf11.setText((new Integer(NbMinImagesPerduesApresPause)).toString());
      tf12.setText((new Integer(NbMaxImagesPerduesApresPause)).toString());
    }
    /** affecte le non des boutons
     */ 
    public void setlabelStart(String nom1){
      bouttonStart.setLabel(nom1);
		}
    
    public void setlabelPause(String nom1){
      bouttonPause.setLabel(nom1);
		}
    
    /** initialise les zones d'affichage
     */ 
    public void init(){
      texteProducteur.setText("");
      texteConsommateur.setText("");
      texteMessages1.setText("");
      texteMessages2.setText("");
      tailleTexte = 0;
    }

    /** Affiche un message dans la fenetre 1 des messages divers
     * @param t le message a afficher
     */ 
    public void afficheMessages1(String t){
      texteMessages1.append(t + "\n");
    }

    /** Affiche un message dans la fenetre 2 des messages divers
     * @param t le message a afficher
     */ 
    public void afficheMessages2(String t){
      tailleTexte = tailleTexte + t.length();
      if (tailleTexte > 75){
        tailleTexte = t.length();
        texteMessages2.append("\n");
      }
      texteMessages2.append(t);
    }
    /** Affiche un message dans la fenetre du Testeur de playout
     * @param t le message a afficher
     */ 
    public void afficheProducteur(String t){
      texteProducteur.append(t + "\n");
    }

    /** Affiche un message dans la fenetre du buffeur de playout
     * @param t le message a afficher
     */ 
    public void afficheConsommateur(String t){
      texteConsommateur.append(t + "\n");
    }
  }
		
		
  /** Objet permettant de simuler l'envoi d'une image sur le reseau
   *  Tiens compte des inversions, pertes et duplication de paquets
   */
  private class EnvoyeurImage extends Thread {
		private VideoImage mvi;
		private boolean duplication = false;
	  private int numRedondance;

    /** Le constructeur
     * @param _mvi L'image a envoyer
     */ 
    public EnvoyeurImage(VideoImage  _mvi, int _numRedondance) {
      mvi = _mvi;
      numRedondance = _numRedondance;			
    }

    /** methode run du thread
     */ 
    public void run() {
      if ((Math.random() * 100) < pourcentagePertes){
        fenetreTest.afficheProducteur(new Integer(mvi.getNum()).toString() + "\t Pas Envoye");
      }
      else {
        if (((Math.random() * 100) < pourcentageDuplication) && (numRedondance > 1)){
          duplication =  true;
          (new EnvoyeurImage(mvi, numRedondance - 1)).start();		
        }
        fenetreTest.afficheMessages2(".");
        try {
          sleep(delaiTransmission + (new Double(Math.random() * variationDelai).longValue()));
        } 
        catch(InterruptedException e) {
          System.err.println("Erreur dans le sleep();");
          e.printStackTrace();
        }
				if (arretDemande == false) {
						if (duplication){
								fenetreTest.afficheProducteur(new Integer(mvi.getNum()).toString() + "\t Dupliquee");
						}
						else{
								fenetreTest.afficheProducteur(new Integer(mvi.getNum()).toString());
						}
						notifye(FrameDisplay.VIDEO_IMAGE, mvi);
						synchronized(EmissionANumSync){
								switch (EmissionANum) {
								case -1 :
										break;
								case 0 :
										fenetreTest.afficheProducteur("\t Emission de donnees a l'image : " + new Integer(mvi.getNum()).toString());
										notifye(FrameDisplay.DELAI_ESTIME,new Long(delaiTransmission));
										notifye(FrameDisplay.VARIATION_DELAI,new Long(variationDelai));
										EmissionANum = EmissionANum -1;
										break;
								default :
										EmissionANum = EmissionANum -1;
										break;
								}
						}
				}
      }
    }
  }





  /** 
   * Methode qui observe un evenement de type PlayoutMessage
   *
   * @see NotifyObject
   * @see VideoImage
   */
  public void update(Observable observable, Object o) {
    NotifyObject no = (NotifyObject)o;
    switch (no.type) {
    case FrameDisplay.PLAYOUT_MESSAGE :
      int niv = ((PlayoutMessage)(no.obj)).getLevel();
      String str = ((PlayoutMessage)(no.obj)).getMessage();
      switch (niv) {
      case 0 :
        System.out.print(str);
        break;
      case 1 :
        fenetreTest.afficheMessages2(str);
        break;
      case 2 :
        fenetreTest.afficheConsommateur(str);
        break;
      case 3 :
        fenetreTest.afficheMessages1(str);
        break;
      default :
        System.out.print(str);        
        break;
      }
    }
  }

  /** methode run du thread
   * @see NotifyObject
   * @see VideoImage
   */
  public void run() {
    fenetreTest.setlabelStart("Arreter");
    notifye(FrameDisplay.INITIALISE_SESSION_VIDEO, null);
    fenetreTest.init();
    fenetreTest.getAff();
    afficheAide();
    long tempsTmp;	
    long tempsRes;
		synchronized(EmissionANumSync){
				EmissionANum = nbImageAvtEmissionInfos - 1;
		}
		notifye(FrameDisplay.GENERE_STATISTIQUES, new Boolean(true));
    notifye(FrameDisplay.COEFFICIENT_VARIATION_DELAI,new Double(CoefficientVariationDelai));
    int i = numPremiereImage;
    while (( i < (nbImagesEnvoyees  + numPremiereImage)) && (arretDemande == false)){
      if (pauseEmission){
        while (pauseEmission){
          try {
            emetteur.yield();
            emetteur.sleep(100);
          } 
          catch(InterruptedException e) {
            System.err.println("Erreur dans le sleep();");
            e.printStackTrace();
          }
        }
        fenetreTest.afficheProducteur("\t Pause apres l'image : " + (i - 1));
        notifye(FrameDisplay.RESET_SESSION_VIDEO, null);
        int NbImagesPerduesApresPause = (int)(Math.random() * (NbMaxImagesPerduesApresPause - NbMinImagesPerduesApresPause)) + NbMinImagesPerduesApresPause;
        i = i + NbImagesPerduesApresPause;
				synchronized(EmissionANumSync){
						EmissionANum = nbImageAvtEmissionInfos - 1;
				}
				if (i < (nbImagesEnvoyees  + numPremiereImage)){
						fenetreTest.afficheProducteur("\t UnPause avant l'image : " + i);
				}
				else{
						fenetreTest.afficheProducteur("\t UnPause ==> fin ");
						break;
				}
      }
			tempsTmp = System.currentTimeMillis();
      vi = new VideoImage(create(i, tailleCoteImage, tailleCoteImage) , i, frequenceEmission, 100);
      Runtime.getRuntime().gc();
      (new EnvoyeurImage(vi, nbMaxDuplication)).start();		
      tempsRes = System.currentTimeMillis();
      if (frequenceEmission >= (tempsRes - tempsTmp)){
        try {
          emetteur.yield();
          emetteur.sleep(frequenceEmission - (tempsRes - tempsTmp));
        } 
        catch(InterruptedException e) {
          System.err.println("Erreur dans le sleep();");
          e.printStackTrace();
        }
      }
      i++;
    }
		synchronized(EmissionANumSync){
				EmissionANum = -1;
		}
  }


  /** notifie le buffer de playout
   * @param o l'objet a notifier
   */
  private void notifye(int type, Object o) {
    setChanged();
    notifyObservers(new NotifyObject(type, o));
  }


  /** cree une image correspondant a un indice
   * @param indice l'indice de l'image
   * @param w largeur de l'image
   * @param h hauteur de l'image
   * @return l'image cree
   */
  private BufferedImage create(int indice, int w, int h){
    BufferedImage bimm = new BufferedImage(w, h, ColorSpace.TYPE_CMYK);
    Graphics2D gc2D = bimm.createGraphics();


    gc2D.setColor(Color.blue);

    int wr = w/7;
    int hr = h/7;

    gc2D.fillOval((indice % w) -(wr/2) ,(indice % h) -(hr/2) , wr, hr);
    gc2D.fillOval((indice % w) -(wr/2),h-(indice % h) -(hr/2), wr, hr);
    gc2D.fillOval(w-(indice % w) -(wr/2),(indice % h) -(hr/2), wr, hr);
    gc2D.fillOval(w-(indice % w) -(wr/2),h-(indice % h) -(hr/2), wr, hr);

    gc2D.setColor(Color.red);
    gc2D.drawLine(0 ,indice % h, w ,indice % h);
    gc2D.drawLine(indice % w, 0,indice % w , h);

    gc2D.drawLine(0 ,h-(indice % h), w ,h-(indice % h));
    gc2D.drawLine(w-(indice % w), 0,w-(indice % w) , h);

    gc2D.setColor(Color.white);
    FontMetrics fontMetrics = gc2D.getFontMetrics();
    String texteAAfficher = (new Integer(indice)).toString();
    Rectangle2D rect = fontMetrics.getStringBounds(texteAAfficher, gc2D); 
    gc2D.drawString(texteAAfficher, w/2 - (int)(rect.getWidth() / 2), h/2 + (int)(rect.getHeight() / 2)); 
    return bimm;
  }
	


  /** Methode main creant le testeur de playout
   */
  public static void main (String argv[]) {
    PlayoutTester p  = new PlayoutTester();
        while(true){
      try {
        Thread.sleep(10000);
      } 
      catch(InterruptedException e) {
       System.err.println("Erreur dans le sleep();");
       e.printStackTrace();
     }
   }
  }
}
